'''
python interface to the Salesforce Metadata API
based on mrbelvedere/mpinstaller/mdapi.py
'''

# TO DO
#   - add docstrings
#   - parse dates from SOAP response
#   - use format() instead of %

import base64
# import dateutil.parser
import httplib
import re
import time
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
from zipfile import ZipFile
import StringIO

import requests

from cumulusci.salesforce_api import soap_envelopes
from cumulusci.core.exceptions import ApexTestException
from cumulusci.utils import zip_subfolder
from cumulusci.salesforce_api.exceptions import MetadataComponentFailure
from cumulusci.salesforce_api.exceptions import MetadataApiError


class BaseMetadataApiCall(object):
    check_interval = 1
    soap_envelope_start = None
    soap_envelope_status = None
    soap_envelope_result = None
    soap_action_start = None
    soap_action_status = None
    soap_action_result = None

    def __init__(self, task):
        # the cumulucci context object contains logger, oauth, ID, secret, etc
        self.task = task
        self.status = None
        self.check_num = 1

    def __call__(self):
        self.task.logger.info('Pending')
        response = self._get_response()
        if self.status != 'Failed':
            return self._process_response(response)

    def _build_endpoint_url(self):
        # Parse org id from id which ends in /ORGID/USERID
        org_id = self.task.org_config.org_id
        # If "My Domain" is configured in the org, the instance_url needs to be
        # parsed differently
        instance_url = self.task.org_config.instance_url
        if instance_url.find('.my.salesforce.com') != -1:
            # Parse instance_url with My Domain configured
            # URL will be in the format
            # https://name--name.na11.my.salesforce.com and should be
            # https://na11.salesforce.com
            instance_url = re.sub(
                r'https://.*\.(\w+)\.my\.salesforce\.com', r'https://\1.salesforce.com', instance_url)
        # Build the endpoint url from the instance_url
        endpoint = '%s/services/Soap/m/33.0/%s' % (instance_url, org_id)
        return endpoint

    def _build_envelope_result(self):
        if self.soap_envelope_result:
            return self.soap_envelope_result % {'process_id': self.process_id}

    def _build_envelope_start(self):
        if self.soap_envelope_start:
            return self.soap_envelope_start

    def _build_envelope_status(self):
        if self.soap_envelope_status:
            return self.soap_envelope_status % {'process_id': self.process_id}

    def _build_headers(self, action, message):
        return {
            'Content-Type': 'text/xml; charset=UTF-8',
            'Content-Length': str(len(message)),
            'SOAPAction': action,
        }

    def _call_mdapi(self, headers, envelope, refresh=None):
        # Insert the session id
        session_id = self.task.org_config.access_token
        auth_envelope = envelope.replace('###SESSION_ID###', session_id)
        response = requests.post(self._build_endpoint_url(
        ), headers=headers, data=auth_envelope)
        faultcode = parseString(
            response.content).getElementsByTagName('faultcode')
        # refresh = False can be passed to prevent a loop if refresh fails
        if refresh is None:
            refresh = True
        if faultcode:
            return self._handle_soap_error(headers, envelope, refresh, response)
        return response

    def _get_element_value(self, dom, tag):
        result = dom.getElementsByTagName(tag)
        if result and result[0].firstChild:
            return result[0].firstChild.nodeValue

    def _get_check_interval(self):
        return self.check_interval * ((self.check_num / 3) + 1)

    def _get_response(self):
        if not self.soap_envelope_start:
            # where is this from?
            raise NotImplemented('No soap_start template was provided')
        # Start the call
        envelope = self._build_envelope_start()
        if not envelope:
            return
        envelope = envelope.encode('utf-8')
        headers = self._build_headers(self.soap_action_start, envelope)
        response = self._call_mdapi(headers, envelope)
        # If no status or result calls are configured, return the result
        if not self.soap_envelope_status and not self.soap_envelope_result:
            return response
        # Process the response to set self.process_id with the process id
        # started
        response = self._process_response_start(response)
        # Check the status if configured
        if self.soap_envelope_status:
            while self.status not in ['Done', 'Failed']:
                # Check status in a loop until done
                envelope = self._build_envelope_status()
                if not envelope:
                    return
                envelope = envelope.encode('utf-8')
                headers = self._build_headers(
                    self.soap_action_status, envelope)
                response = self._call_mdapi(headers, envelope)
                response = self._process_response_status(response)

                # start increasing the check interval progressively to handle long pending jobs
                check_interval = self._get_check_interval()
                self.check_num += 1

                time.sleep(check_interval)
            # Fetch the final result and return
            if self.soap_envelope_result:
                envelope = self._build_envelope_result()
                if not envelope:
                    return
                envelope = envelope.encode('utf-8')
                headers = self._build_headers(
                    self.soap_action_result, envelope)
                response = self._call_mdapi(headers, envelope)
            else:
                return response
        else:
            # Check the result and return when done
            while self.status not in ['Succeeded', 'Failed', 'Cancelled']:
                # start increasing the check interval progressively to handle long pending jobs
                check_interval = self._get_check_interval()
                self.check_num += 1
                time.sleep(check_interval)

                envelope = self._build_envelope_result()
                envelope = envelope.encode('utf-8')
                headers = self._build_headers(
                    self.soap_action_result, envelope)
                response = self._call_mdapi(headers, envelope)
                response = _process_response_result(response)
        return response

    def _handle_soap_error(self, headers, envelope, refresh, response):
        faultcode = parseString(
            response.content).getElementsByTagName('faultcode')
        if faultcode:
            faultcode = faultcode[0].firstChild.nodeValue
        else:
            faultcode = ''
        faultstring = parseString(
            response.content).getElementsByTagName('faultstring')
        if faultstring:
            faultstring = faultstring[0].firstChild.nodeValue
        else:
            faultstring = response.content
        if faultcode == 'sf:INVALID_SESSION_ID' and self.task.org_config and self.task.org_config.refresh_token:
            # Attempt to refresh token and recall request
            if refresh:
                self.org_config.refresh_oauth_token()
                return self._call_mdapi(headers, envelope, refresh=False)
        # Log the error
        message = '{}: {}'.format(faultcode, faultstring)
        self._set_status('Failed', message)
        raise MetadataApiError(message, response)

    def _process_response(self, response):
        return response

    def _process_response_result(self, response):
        self._set_status('Succeeded')
        return response

    def _process_response_start(self, response):
        if response.status_code == httplib.INTERNAL_SERVER_ERROR:
            return response
        ids = parseString(response.content).getElementsByTagName('id')
        if ids:
            self.process_id = ids[0].firstChild.nodeValue
        return response

    def _process_response_status(self, response):
        resp_xml = parseString(response.content)
        done = resp_xml.getElementsByTagName('done')
        if done:
            if done[0].firstChild.nodeValue == 'true':
                self._set_status('Done')
            else:
                state_detail = resp_xml.getElementsByTagName('stateDetail')
                if state_detail:
                    log = state_detail[0].firstChild.nodeValue
                    self._set_status('InProgress', log)
                    self.check_num = 1
                elif self.status == 'InProgress':
                    self.check_num = 1
                    self._set_status('InProgress', 'next check in {} seconds'.format(self._get_check_interval()))
                else:
                    self._set_status('Pending', 'next check in {} seconds'.format(self._get_check_interval()))
        else:
            # If no done element was in the xml, fail logging the entire SOAP
            # envelope as the log
            self._set_status('Failed', response.content, response=response)
        return response

    def _set_status(self, status, log=None, level=None, response=None):
        if not level:
            level = 'info'
            if status == 'Failed':
                level = 'error'
        logger = getattr(self.task.logger, level)
        self.status = status
        if log:
            logger('[{}]: {}'.format(status, log))
        else:
            logger('[{}]'.format(status))


class ApiRetrieveUnpackaged(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_UNPACKAGED
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = 'retrieve'
    soap_action_status = 'checkStatus'
    soap_action_result = 'checkRetrieveStatus'

    def __init__(self, task, package_xml, api_version):
        super(ApiRetrieveUnpackaged, self).__init__(task)
        self.package_xml = package_xml
        self.api_version = api_version
        self._clean_package_xml()

    def _clean_package_xml(self):
        self.package_xml = re.sub('<\?xml.*\?>', '', self.package_xml)
        self.package_xml = re.sub('<Package.*>', '', self.package_xml, 1)
        self.package_xml = re.sub('</Package>', '', self.package_xml, 1)
        self.package_xml = re.sub('\n', '', self.package_xml)
        self.package_xml = re.sub(' +<', '<', self.package_xml)

    def _build_envelope_start(self):
        return self.soap_envelope_start.format(
            self.api_version,
            self.package_xml,
        )

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName('zipFile')
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        else:
            return self.packages
        zipstringio = StringIO.StringIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, 'r')
        zipfile = zip_subfolder(zipfile, 'unpackaged')
        return zipfile

class ApiRetrieveInstalledPackages(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_INSTALLEDPACKAGE
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = 'retrieve'
    soap_action_status = 'checkStatus'
    soap_action_result = 'checkRetrieveStatus'

    def __init__(self, task):
        super(ApiRetrieveInstalledPackages, self).__init__(task)
        self.packages = []

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName('zipFile')
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        else:
            return self.packages
        zipstringio = StringIO.StringIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, 'r')
        packages = {}
        # Loop through all files in the zip skipping anything other than
        # InstalledPackages
        for path in zipfile.namelist():
            if not path.endswith('.installedPackage'):
                continue
            namespace = path.split('/')[-1].split('.')[0]
            version = parseString(zipfile.open(
                path).read()).getElementsByTagName('versionNumber')
            if version:
                version = version[0].firstChild.nodeValue
            packages[namespace] = version
        self.packages = packages
        return self.packages


class ApiRetrievePackaged(BaseMetadataApiCall):
    check_interval = 1
    soap_envelope_start = soap_envelopes.RETRIEVE_PACKAGED
    soap_envelope_status = soap_envelopes.CHECK_STATUS
    soap_envelope_result = soap_envelopes.CHECK_RETRIEVE_STATUS
    soap_action_start = 'retrieve'
    soap_action_status = 'checkStatus'
    soap_action_result = 'checkRetrieveStatus'

    def __init__(self, task, package_name, api_version):
        super(ApiRetrievePackaged, self).__init__(task)
        self.package_name = package_name
        self.api_version = api_version

    def _build_envelope_start(self):
        return self.soap_envelope_start.format(
            self.api_version,
            escape(self.package_name),
        )

    def _process_response(self, response):
        # Parse the metadata zip file from the response
        zipstr = parseString(response.content).getElementsByTagName('zipFile')
        if zipstr:
            zipstr = zipstr[0].firstChild.nodeValue
        else:
            return self.packages
        zipstringio = StringIO.StringIO(base64.b64decode(zipstr))
        zipfile = ZipFile(zipstringio, 'r')
        return zipfile


class ApiDeploy(BaseMetadataApiCall):
    soap_envelope_start = soap_envelopes.DEPLOY
    soap_envelope_status = soap_envelopes.CHECK_DEPLOY_STATUS
    soap_action_start = 'deploy'
    soap_action_status = 'checkDeployStatus'

    def __init__(self, task, package_zip, purge_on_delete=None):
        super(ApiDeploy, self).__init__(task)
        if purge_on_delete is None:
            purge_on_delete = True
        self._set_purge_on_delete(purge_on_delete)
        self.package_zip = package_zip

    def _set_purge_on_delete(self, purge_on_delete):
        if purge_on_delete == False or purge_on_delete == 'false':
            self.purge_on_delete = 'false'
        else:
            self.purge_on_delete = 'true'
        # Disable purge on delete entirely for non sandbox or DE orgs as it is
        # not allowed
        # FIXME: To implement this, the task needs to be able to provide the org_type
        #org_type = self.task.org_config.org_type
        #if org_type.find('Sandbox') == -1 and org_type != 'Developer Edition':
        #    self.purge_on_delete = 'false'

    def _build_envelope_start(self):
        if self.package_zip:
            return self.soap_envelope_start % {
                'package_zip': self.package_zip,
                'purge_on_delete': self.purge_on_delete,
            }

    def _process_response(self, response):
        status = parseString(response.content).getElementsByTagName('status')
        if status:
            status = status[0].firstChild.nodeValue
        else:
            # If no status element is in the result xml, return fail and log
            # the entire SOAP envelope in the log
            self._set_status('Failed', response.content)
            return self.status
        # Only done responses should be passed so we need to handle any status
        # related to done
        if status in ['Succeeded', 'SucceededPartial']:
            self._set_status('Success', status)
        else:
            # If failed, parse out the problem text and raise appropriate exception
            messages = []
            resp_xml = parseString(response.content)

            component_failures = resp_xml.getElementsByTagName('componentFailures')
            for component_failure in component_failures:
                failure_info = {
                    'component_type': None,
                    'file_name': None,
                    'line_num': None,
                    'column_num': None,
                    'problem': component_failure.getElementsByTagName('problem')[0].firstChild.nodeValue,
                    'problem_type': component_failure.getElementsByTagName('problemType')[0].firstChild.nodeValue,
                }
                component_type = component_failure.getElementsByTagName('componentType')
                if component_type and component_type[0].firstChild:
                    failure_info['component_type'] = component_type[0].firstChild.nodeValue
                file_name = component_failure.getElementsByTagName('fullName')
                if file_name and file_name[0].firstChild:
                    failure_info['file_name'] = file_name[0].firstChild.nodeValue
                if not failure_info['file_name']:
                    file_name = component_failure.getElementsByTagName('fileName')
                    if file_name and file_name[0].firstChild:
                        failure_info['file_name'] = file_name[0].firstChild.nodeValue
        
                line_num = component_failure.getElementsByTagName('lineNumber')
                if line_num and line_num[0].firstChild:
                    failure_info['line_num'] = line_num[0].firstChild.nodeValue
                
                column_num = component_failure.getElementsByTagName('columnNumber')
                if column_num and column_num[0].firstChild:
                    failure_info['column_num'] = column_num[0].firstChild.nodeValue
                
                created = component_failure.getElementsByTagName('created')[0].firstChild.nodeValue == 'true'
                deleted = component_failure.getElementsByTagName('deleted')[0].firstChild.nodeValue == 'true'
                if deleted: 
                    failure_info['action'] = 'Delete'
                elif created:
                    failure_info['action'] = 'Create'
                else:
                    failure_info['action'] = 'Update'
  
                if failure_info['file_name'] and failure_info['line_num']: 
                    messages.append('{action} of {component_type} {file_name}: {problem_type} on line {line_num}, col {column_num}: {problem}'.format(**failure_info))
                elif failure_info['file_name']:
                    messages.append('{action} of {component_type} {file_name}: {problem_type}: {problem}'.format(**failure_info))
                else:
                    messages.append('{action} of {problem_type}: {problem}'.format(**failure_info))

            if messages:
                # Deploy failures due to a component failure should raise MetadataComponentFailure
                log = '\n\n'.join(messages)
                self._set_status('Failed', log)
                raise MetadataComponentFailure(log, response)
                
            else:
                problems = parseString(
                    response.content).getElementsByTagName('problem')
                for problem in problems:
                    messages.append(problem.firstChild.nodeValue)

            # Parse out any failure text (from test failures in production
            # deployments) and add to log
            failures = parseString(
                response.content).getElementsByTagName('failures')
            for failure in failures:
                # Get needed values from subelements
                namespace = failure.getElementsByTagName('namespace')
                if namespace and namespace[0].firstChild:
                    namespace = namespace[0].firstChild.nodeValue
                else:
                    namespace = None
                stacktrace = failure.getElementsByTagName('stackTrace')
                if stacktrace and stacktrace[0].firstChild:
                    stacktrace = stacktrace[0].firstChild.nodeValue
                else:
                    stacktrace = None
                message = ['Apex Test Failure: ', ]
                if namespace:
                    message.append('from namespace %s: ' % namespace)
                if stacktrace:
                    message.append(stacktrace)
                messages.append(''.join(message))
            if messages:
                log = '\n\n'.join(messages)
            else:
                log = response.content

            if messages:
                # Deploy failures due to a component failure should raise MetadataComponentFailure
                log = '\n\n'.join(messages)
                self._set_status('Failed', log)
                raise ApexTestException(log)

            self._set_status('Failed', log)
            raise MetadataApiError(log, response)

        return self.status


class ApiInstallVersion(ApiDeploy):

    def __init__(self, task, version, purge_on_delete=False):
        self.version = version
        # Construct and set the package_zip file
        if self.version.number:
            self.package_zip = PackageZipBuilder(
                self.version.package.namespace, self.version.number).install_package()
        elif self.version.zip_url or self.version.repo_url:
            if self.version.repo_url:
                repo_url = self.version.repo_url
                git_ref = self.version.branch
                if installation_step.installation.git_ref:
                    git_ref = installation_step.installation.git_ref
                if installation_step.installation.fork:
                    repo_url_parts = repo_url.split('/')
                    repo_url_parts[3] = installation_step.installation.fork
                    repo_url = '/'.join(repo_url_parts)
                zip_url = '%s/archive/%s.zip' % (repo_url, git_ref)
            else:
                zip_url = self.version.zip_url
            # Deploy a zipped bundled downloaded from a URL
            try:
                zip_resp = requests.get(zip_url)
            except:
                raise ValueError('Failed to fetch zip from %s' %
                                 self.version.zip_url)
            zipfp = StringIO.StringIO(zip_resp.content)
            zipfile = ZipFile(zipfp, 'r')
            if not self.version.subfolder and not self.version.repo_url:
                zipfile.close()
                zipfp.seek(0)
                self.package_zip = base64.b64encode(zipfp.read())
            else:
                ignore_prefix = ''
                if self.version.repo_url:
                    # Get the top level folder from the zip
                    ignore_prefix = '%s/' % zipfile.namelist()[0].split('/')[0]
                # Extract a subdirectory from the zip
                subdirectory = ignore_prefix + self.version.subfolder
                subzip = zip_subfolder(
                    zipfile, subdirectory, self.version.namespace_token, self.version.namespace)
                subzipfp = subzip.fp
                subzip.close()
                subzipfp.seek(0)
                self.package_zip = base64.b64encode(subzipfp.read())
        super(ApiInstallVersion, self).__init__(
            task, self.package_zip, purge_on_delete)


class ApiUninstallVersion(ApiDeploy):

    def __init__(self, task, version, purge_on_delete=True):
        self.version = version
        if not version.number:
            self.package_zip = None
        else:
            self.package_zip = PackageZipBuilder(
                self.version.package.namespace).uninstall_package()
        super(ApiUninstallVersion, self).__init__(
            task, self.package_zip, purge_on_delete)


class ApiListMetadata(BaseMetadataApiCall):
    soap_envelope_start = soap_envelopes.LIST_METADATA
    soap_action_start = 'listMetadata'

    def __init__(self, task, metadata_type, metadata=None, folder=None):
        super(ApiListMetadata, self).__init__(task)
        self.metadata_type = metadata_type
        self.metadata = metadata
        self.folder = folder
        if self.metadata is None:
            self.metadata = {}

    def _build_envelope_start(self):
        folder = self.folder
        if folder is None:
            folder = ''
        else:
            folder = '\n      <folder>{}</folder>'.format(folder)      
        return self.soap_envelope_start % {'metadata_type': self.metadata_type, 'folder': self.folder}

    def _process_response(self, response):
        metadata = []
        tags = [
            'createdById',
            'createdByName',
            'createdDate',
            'fileName',
            'fullName',
            'id',
            'lastModifiedById',
            'lastModifiedByName',
            'lastModifiedDate',
            'manageableState',
            'namespacePrefix',
            'type',
        ]
        # These tags will be interpreted into dates
        parse_dates = [
            'createdDate',
            'lastModifiedDate',
        ]
        for result in parseString(response.content).getElementsByTagName('result'):
            result_data = {}
            # Parse fields
            for tag in tags:
                result_data[tag] = self._get_element_value(result, tag)
            # Parse dates
            # FIXME: This was breaking things
            # for key in parse_dates:
            #    if result_data[key]:
            #        result_data[key] = dateutil.parser.parse(result_data[key])
            metadata.append(result_data)
        if self.metadata_type in self.metadata:
            self.metadata[self.metadata_type].extend(metadata)
        else:
            self.metadata[self.metadata_type] = metadata
        return self.metadata



""" Tests for the CumulusCI task module """

import unittest
import logging
import collections

from cumulusci.core.tasks import BaseTask
from cumulusci.core.flows import BaseFlow
from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import FlowConfig
from cumulusci.core.tests.utils import MockLoggingHandler
import cumulusci.core

ORG_ID = '00D000000000001'
USERNAME = 'sample@example'

class _TaskHasResult(BaseTask):
    def _run_task(self):
        return -1


class _SfdcTask(BaseTask):
    salesforce_task = True

    def _run_task(self):
        return -1


class TestBaseTaskCallable(unittest.TestCase):
    """ Tests for the BaseTask callable interface.

    BaseTask is a callable interface
    BaseTask has return_values and results
    BaseTask has basic logging
    """

    task_class = BaseTask

    @classmethod
    def setUpClass(cls):
        super(TestBaseTaskCallable, cls).setUpClass()
        logger = logging.getLogger(cumulusci.core.tasks.__name__)
        logger.setLevel(logging.DEBUG)
        cls._task_log_handler = MockLoggingHandler(logging.DEBUG)
        logger.addHandler(cls._task_log_handler)

    def setUp(self):
        self.global_config = BaseGlobalConfig()
        self.project_config = BaseProjectConfig(self.global_config)
        self.org_config = OrgConfig({
            'username': USERNAME,
            'org_id': ORG_ID
        })
        self.task_config = TaskConfig()
        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_task_is_callable(self):
        """ BaseTask is Callable """
        task = self.__class__.task_class(
            self.project_config,
            self.task_config,
            self.org_config
        )

        self.assertIsInstance(task, collections.Callable)

    def test_get_return_values(self):
        """ Callable interface returns retvals """

        class _TaskReturnsStuff(BaseTask):
            def _run_task(self):
                self.return_values['name'] = 'return!'

        task = _TaskReturnsStuff(
            self.project_config,
            self.task_config,
            self.org_config
        )
        return_values = task()

        self.assertIn('name', return_values)

    def test_get_task_result(self):
        """ Task results available as an instance member """

        task = _TaskHasResult(
            self.project_config,
            self.task_config,
            self.org_config
        )
        task()

        self.assertEqual(task.result, -1)

    def test_task_logs_name_not_org(self):
        """ A task logs the task class name to info (and not creds) """

        task = _TaskHasResult(
            self.project_config,
            self.task_config,
            self.org_config
        )
        task()

        self.assertTrue(any(
            "_TaskHasResult" in s for s in self.task_log['info']
        ))

        self.assertFalse(any(
            ORG_ID in s for s in self.task_log['info']
        ))

    def test_salesforce_task_logs_org_id(self):
        """ A salesforce_task will also log the org id & username """

        task = _SfdcTask(
            self.project_config,
            self.task_config,
            self.org_config
        )
        task()

        self.assertTrue(any(
            ORG_ID in s for s in self.task_log['info']
        ))

    def test_no_id_if_run_from_flow(self):
        """ A salesforce_task will not log the org id if run from a flow """

        task = _SfdcTask(
            self.project_config,
            self.task_config,
            self.org_config,
            flow=BaseFlow(self.project_config, FlowConfig(), self.org_config)
        )
        task()
        self.assertFalse(any(
            ORG_ID in s for s in self.task_log['info']
        ))



class CumulusCIException(Exception):
    pass

class NotInProject(CumulusCIException):
    """ Raised when no project can be found in the current context """
    pass

class ProjectConfigNotFound(CumulusCIException):
    """ Raised when a project is found in the current context but no configuration was found for the project """
    pass

class KeychainNotFound(CumulusCIException):
    """ Raised when no keychain could be found """
    pass

class KeychainKeyNotFound(CumulusCIException):
    """ Raised when the keychain key couldn't be found """

class KeychainConnectedAppNotFound(CumulusCIException):
    """ Raised when the connected app configuration couldn't be found for a keychain """

class OrgNotFound(CumulusCIException):
    """ Raised when no org could be found by a given name in the project keychain """
    pass

class ServiceNotConfigured(CumulusCIException):
    """ Raised when no service configuration could be found by a given name in the project keychain """
    pass

class ServiceNotValid(CumulusCIException):
    """ Raised when no service configuration could be found by a given name in the project configuration """
    pass

class ConfigError(CumulusCIException):
    """ Raised when a configuration enounters an error """
    pass

class AntTargetException(CumulusCIException):
    """ Raised when a generic Ant target error occurs """
    pass

class DeploymentException(CumulusCIException):
    """ Raised when a metadata api deployment error occurs """
    pass

class ApexTestException(CumulusCIException):
    """ Raised when a build fails because of an Apex test failure """
    pass

class SalesforceCredentialsException(CumulusCIException):
    """ Raise when Salesforce credentials are invalid """
    pass

class TaskRequiresSalesforceOrg(CumulusCIException):
    """ Raise when a task that requires a Salesforce org_config is not initialized with an org_config """
    pass

class TaskOptionsError(CumulusCIException):
    """ Raise when a task's options are invalid """
    pass

class GithubNotConfigured(CumulusCIException):
    """ Raise when attempting to get the Github configuration from the keychain and no configuration is set """
    pass

class MrbelvedereNotConfigured(CumulusCIException):
    """ Raise when attempting to get the mrbelvedere configuration from the keychain and no configuration is set """
    pass

class ApexTestsDBNotConfigured(CumulusCIException):
    """ Raise when attempting to get the ApexTestsDB configuration from the keychain and no configuration is set """
    pass

class TaskNotFoundError(CumulusCIException):
    """ Raise when task is not found in project config """
    pass

class FlowNotFoundError(CumulusCIException):
    """ Raise when flow is not found in project config """
    pass

class MrbelvedereError(CumulusCIException):
    """ Raise for errors from mrbelvedere installer """
    pass

class ApextestsdbError(CumulusCIException):
    """ Raise for errors from apextestsdb """
    pass

class ScratchOrgException(CumulusCIException):
    """ Raise for errors related to scratch orgs """
    pass

class GithubException(CumulusCIException):
    """ Raise for errors related to GitHub """
    pass

class SalesforceException(CumulusCIException):
    """ Raise for errors related to Salesforce """
    pass

class SalesforceDXException(CumulusCIException):
    """ Raise for errors related to Salesforce DX """
    pass

class SOQLQueryException(CumulusCIException):
    """ Raise for errors related to Salesforce DX """
    pass

class CommandException(CumulusCIException):
    """ Raise for errors coming from spawned CLI subprocesses """
    pass



""" Tasks are the basic unit of execution in CumulusCI.

Subclass BaseTask or a descendant to define custom task logic
"""

import logging
import time

from cumulusci.core.exceptions import TaskRequiresSalesforceOrg
from cumulusci.core.exceptions import TaskOptionsError


class BaseTask(object):
    """ BaseTask provides the core execution logic for a Task

    Subclass BaseTask and provide a `_run_task()` method with your
    code.
    """
    task_options = {}
    salesforce_task = False  # Does this task require a salesforce org?

    def __init__(self, project_config, task_config,
                 org_config=None, flow=None, **kwargs):
        self.project_config = project_config
        self.task_config = task_config
        self.org_config = org_config
        self.return_values = {}
        """ a dict of return_values that can be used by task callers """
        self.result = None
        """ a simple result object for introspection, often a return_code """
        self.flow = flow
        """ The flow for this task execution """
        if self.salesforce_task and not self.org_config:
            raise TaskRequiresSalesforceOrg('This task requires a Saleforce '
                                            'org_config but none was passed '
                                            'to the Task constructor')
        self._init_logger()
        self._init_options(kwargs)
        self._validate_options()
        self._update_credentials()
        self._init_task()

    def _init_logger(self):
        """ Initializes self.logger """
        self.logger = logging.getLogger(__name__)

    def _init_options(self, kwargs):
        """ Initializes self.options """
        self.options = self.task_config.options
        if self.options is None:
            self.options = {}
        if kwargs:
            self.options.update(kwargs)

    def _validate_options(self):
        missing_required = []
        for name, config in self.task_options.items():
            if config.get('required') is True and name not in self.options:
                missing_required.append(name)

        if missing_required:
            raise TaskOptionsError(
                '{} requires the options ({}) '
                'and no values were provided'.format(
                    self.__class__.__name__,
                    ', '.join(missing_required)
                )
            )

    def _update_credentials(self):
        """ Override to do any logic  to refresh credentials """
        pass

    def _init_task(self):
        """ Override to implement dynamic logic for initializing the task. """
        pass

    def __call__(self):
        self._log_begin()
        self.result = self._run_task()
        return self.return_values

    def _run_task(self):
        """ Subclasses should override to provide their implementation """
        pass

    def _log_begin(self):
        """ Log the beginning of the task execution """
        self.logger.info('Beginning task: %s', self.__class__.__name__)
        if self.salesforce_task and not self.flow:
            self.logger.info('%15s %s', 'As user:', self.org_config.username)
            self.logger.info('%15s %s', 'In org:', self.org_config.org_id)
        self.logger.info('')

    def _retry(self):
        while True:
            try:
                self._try()
                break
            except Exception as e:
                if not (self.options['retries'] and self._is_retry_valid(e)):
                    raise
                if self.options['retry_interval']:
                    self.logger.warning(
                        'Sleeping for {} seconds before retry...'.format(
                        self.options['retry_interval']))
                    time.sleep(self.options['retry_interval'])
                    if self.options['retry_interval_add']:
                        self.options['retry_interval'] += self.options[
                            'retry_interval_add']
                self.options['retries'] -= 1
                self.logger.warning(
                    'Retrying ({} attempts remaining)'.format(
                    self.options['retries']))

    def _try(self):
        raise NotImplementedError(
            'Subclasses should provide their own implementation')

    def _is_retry_valid(self, e):
        return True



""" The flow engine allows a series of tasks to be run. """

import copy
from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module
import logging
import traceback

from cumulusci.core.config import TaskConfig
from cumulusci.core.utils import import_class


class BaseFlow(object):
    """ BaseFlow handles initializing and running a flow """
    def __init__(self, project_config, flow_config, org_config):
        self.project_config = project_config
        self.flow_config = flow_config
        self.org_config = org_config
        self.task_return_values = []
        """ A collection of return_values dicts in task execution order """
        self.task_results = []
        """ A collection of result objects in task execution order """
        self.tasks = []
        """ A collection of configured task objects, either run or failed """
        self._init_logger()
        self._init_flow()

    def _init_logger(self):
        """ Initializes self.logger """
        self.logger = logging.getLogger(__name__)

    def _init_flow(self):
        """ Initialize the flow and print flow details to info """
        self.logger.info('---------------------------------------')
        self.logger.info(
            'Initializing flow class %s:',
            self.__class__.__name__,
        )
        self.logger.info('---------------------------------------')
        for line in self._render_config():
            self.logger.info(line)

    def _get_tasks(self):
        tasks = []
        for step_num, config in self.flow_config.tasks.items():
            if config['task'] == 'None':
                continue
            tasks.append((
                LooseVersion(str(step_num)),
                {
                    'flow_config': config,
                    'task_config': self.project_config.get_task(
                        config['task']
                    ),
                }
            ))
        tasks.sort()
        tasks = [task_info[1] for task_info in tasks]
        return tasks

    def _render_config(self):
        config = []
        config.append(
            'Flow Description: {}'.format(self.flow_config.description)
        )

        if not self.flow_config.tasks:
            return config

        config.append('Tasks:')
        for task_info in self._get_tasks():
            config.append('  {}: {}'.format(
                task_info['flow_config']['task'],
                task_info['task_config'].description,
            ))

        if self.org_config is not None:
            config.append('Organization:')
            config.append('  {}: {}'.format('Username', self.org_config.username))
            config.append('  {}: {}'.format('  Org Id', self.org_config.org_id))

        return config

    def __call__(self):
        for flow_task_config in self._get_tasks():
            self._run_task(flow_task_config)

    def _find_task_by_name(self, name):
        if not self.flow_config.tasks:
            return

        i = 0
        for task in self._get_tasks():
            if task['flow_config']['task'] == name:
                if len(self.tasks) > i:
                    return self.tasks[i]
            i += 1

    def _run_task(self, flow_task_config):
        task_config = copy.deepcopy(flow_task_config['task_config'].config)
        task_config = TaskConfig(task_config)

        task_name = flow_task_config['flow_config']['task']

        if 'options' not in task_config.config:
            task_config.config['options'] = {}
        task_config.config['options'].update(
            flow_task_config['flow_config'].get('options', {})
        )

        # Handle dynamic value lookups in the format ^^task_name.attr1.attr2
        for option, value in task_config.options.items():
            if unicode(value).startswith('^^'):
                value_parts = value[2:].split('.')
                parent = self._find_task_by_name(value_parts[0])
                for attr in value_parts[1:]:
                    parent = parent.return_values.get(attr)
                task_config.config['options'][option] = parent

        task_class = import_class(task_config.class_path)

        self.logger.info('')
        self.logger.info('Running task: %s', task_name)

        task = task_class(
            self.project_config,
            task_config,
            org_config=self.org_config,
            flow=self
        )
        self.tasks.append(task)

        for line in self._render_task_config(task):
            self.logger.info(line)

        try:
            task()
            self.logger.info('Task complete: %s', task_name)
            self.task_results.append(task.result)
            self.task_return_values.append(task.return_values)
        except Exception as e:
            self.logger.error('Task failed: %s', task_name)
            if not flow_task_config['flow_config'].get('ignore_failure'):
                self.logger.error('Failing flow due to exception in task')
                traceback.print_exc()
                raise e
            self.logger.info('Continuing flow')

    def _render_task_config(self, task):
        config = ['Options:']
        if not task.task_options:
            return config

        for option_name in list(task.task_options.keys()):
            config.append('  {}: {}'.format(
                option_name,
                task.options.get(option_name),
            ))

        return config



import httplib
import re
import os
import requests
import json

from datetime import datetime
from distutils.version import LooseVersion

from cumulusci.tasks.release_notes.github_api import GithubApiMixin
from cumulusci.tasks.release_notes.parser import ChangeNotesLinesParser
from cumulusci.tasks.release_notes.parser import IssuesParser
from cumulusci.tasks.release_notes.parser import GithubIssuesParser
from cumulusci.tasks.release_notes.parser import GithubLinesParser
from cumulusci.tasks.release_notes.parser import GithubLinkingLinesParser
from cumulusci.tasks.release_notes.parser import CommentingGithubIssuesParser
from cumulusci.tasks.release_notes.provider import StaticChangeNotesProvider
from cumulusci.tasks.release_notes.provider import DirectoryChangeNotesProvider
from cumulusci.tasks.release_notes.provider import GithubChangeNotesProvider
from cumulusci.tasks.release_notes.exceptions import GithubApiNotFoundError


class BaseReleaseNotesGenerator(object):

    def __init__(self):
        self.change_notes = []
        self.init_parsers()
        self.init_change_notes()

    def __call__(self):
        self._parse_change_notes()
        return self.render()

    def init_change_notes(self):
        self.change_notes = self._init_change_notes()

    def _init_change_notes(self):
        """ Subclasses should override this method to return an initialized
        subclass of BaseChangeNotesProvider """
        return []

    def init_parsers(self):
        """ Initializes the parser instances as the list self.parsers """
        self.parsers = []
        self._init_parsers()

    def _init_parsers(self):
        """ Subclasses should override this method to initialize their
        parsers """
        pass

    def _parse_change_notes(self):
        """ Parses all change_notes in self.change_notes() through all parsers
        in self.parsers """
        for change_note in self.change_notes():
            self._parse_change_note(change_note)

    def _parse_change_note(self, change_note):
        """ Parses an individual change note through all parsers in
        self.parsers """
        for parser in self.parsers:
            parser.parse(change_note)

    def render(self):
        """ Returns the rendered release notes from all parsers as a string """
        release_notes = []
        for parser in self.parsers:
            parser_content = parser.render()
            if parser_content is not None:
                release_notes.append(parser_content)
        return u'\r\n\r\n'.join(release_notes)


class StaticReleaseNotesGenerator(BaseReleaseNotesGenerator):

    def __init__(self, change_notes):
        self._change_notes = change_notes
        super(StaticReleaseNotesGenerator, self).__init__()

    def _init_parsers(self):
        self.parsers.append(ChangeNotesLinesParser(
            self, 'Critical Changes'))
        self.parsers.append(ChangeNotesLinesParser(self, 'Changes'))
        self.parsers.append(IssuesParser(
            self, 'Issues Closed'))

    def _init_change_notes(self):
        return StaticChangeNotesProvider(self, self._change_notes)


class DirectoryReleaseNotesGenerator(BaseReleaseNotesGenerator):

    def __init__(self, directory):
        self.directory = directory
        super(DirectoryReleaseNotesGenerator, self).__init__()

    def _init_parsers(self):
        self.parsers.append(ChangeNotesLinesParser(
            self, 'Critical Changes'))
        self.parsers.append(ChangeNotesLinesParser(self, 'Changes'))
        self.parsers.append(IssuesParser(
            self, 'Issues Closed'))

    def _init_change_notes(self):
        return DirectoryChangeNotesProvider(self, self.directory)


class GithubReleaseNotesGenerator(BaseReleaseNotesGenerator):

    def __init__(self, github_info, current_tag, last_tag=None, link_pr=False):
        self.github_info = github_info
        self.current_tag = current_tag
        self.last_tag = last_tag
        self.link_pr = link_pr
        self.lines_parser_class = None
        self.issues_parser_class = None
        super(GithubReleaseNotesGenerator, self).__init__()

    def _init_parsers(self):
        self._set_classes()
        self.parsers.append(self.lines_parser_class(
            self,
            'Critical Changes',
        ))
        self.parsers.append(self.lines_parser_class(
            self,
            'Changes',
        ))
        self.parsers.append(self.issues_parser_class(
            self,
            'Issues Closed',
            link_pr=self.link_pr,
        ))

    def _init_change_notes(self):
        return GithubChangeNotesProvider(
            self,
            self.current_tag,
            self.last_tag
        )

    def _set_classes(self):
        self.lines_parser_class = (GithubLinkingLinesParser if self.link_pr
                            else GithubLinesParser)
        self.issues_parser_class = GithubIssuesParser


class PublishingGithubReleaseNotesGenerator(GithubReleaseNotesGenerator, GithubApiMixin):

    def __call__(self):
        content = super(PublishingGithubReleaseNotesGenerator, self).__call__()
        return self.publish(content)

    def publish(self, content):
        release = self._get_release()
        return self._update_release(release, content)

    def _get_release(self):
        # Query for the release
        return self.call_api('/releases/tags/{}'.format(self.current_tag))

    def _set_classes(self):
        self.lines_parser_class = (GithubLinkingLinesParser if self.link_pr
                            else GithubLinesParser)
        self.issues_parser_class = CommentingGithubIssuesParser

    def _update_release(self, release, content):

        if release['body']:
            new_body = []
            current_parser = None
            is_start_line = False
            for parser in self.parsers:
                parser.replaced = False

            # update existing sections
            for line in release['body'].splitlines():

                if current_parser:
                    if current_parser._is_end_line(current_parser._process_line(line)):
                        parser_content = current_parser.render()
                        if parser_content:
                            # replace existing section with new content
                            new_body.append(parser_content + '\r\n')
                        current_parser = None

                for parser in self.parsers:
                    if parser._render_header().strip() == parser._process_line(line).strip():
                        parser.replaced = True
                        current_parser = parser
                        is_start_line = True
                        break
                    else:
                        is_start_line = False

                if is_start_line:
                    continue
                if current_parser:
                    continue
                else:
                    # preserve existing sections
                    new_body.append(line.strip())

            # catch section without end line
            if current_parser:
                new_body.append(current_parser.render())

            # add new sections at bottom
            for parser in self.parsers:
                parser_content = parser.render()
                if parser_content and not parser.replaced:
                    new_body.append(parser_content + '\r\n')

            release['body'] = u'\r\n'.join(new_body)
        else:
            release['body'] = content

        if release.get('id'):
            resp = self.call_api(
                '/releases/{}'.format(release['id']), data=release)
        else:
            resp = self.call_api('/releases', data=release)

        return release['body']
